@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Implementation Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -38,6 +38,7 @@ static void gail_window_class_init (GailWindowClass *klass);
 
 static void                  gail_window_real_initialize (AtkObject    *obj,
                                                           gpointer     data);
+static void                  gail_window_finalize        (GObject      *object);
 
 static G_CONST_RETURN gchar* gail_window_get_name       (AtkObject     *accessible);
 
@@ -112,8 +113,11 @@ static void
 gail_window_class_init (GailWindowClass *klass)
 {
   GailWidgetClass *widget_class;
+  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
   AtkObjectClass  *class = ATK_OBJECT_CLASS (klass);
 
+  gobject_class->finalize = gail_window_finalize;
+
   widget_class = (GailWidgetClass*)klass;
   widget_class->focus_gtk = gail_window_real_focus_gtk;
   widget_class->notify_gtk = gail_window_real_notify_gtk;
@@ -211,29 +215,12 @@ gail_window_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-
-  if (GTK_IS_FILE_SELECTION (widget))
-    accessible->role = ATK_ROLE_FILE_CHOOSER;
-  else if (GTK_IS_COLOR_SELECTION_DIALOG (widget))
-    accessible->role = ATK_ROLE_COLOR_CHOOSER;
-  else if (GTK_IS_FONT_SELECTION_DIALOG (widget))
-    accessible->role = ATK_ROLE_FONT_CHOOSER;
-  else if (GTK_IS_DIALOG (widget))
-    accessible->role = ATK_ROLE_DIALOG;
-  else if (GTK_IS_WINDOW (widget))
-    {
-      const gchar *name;
-
-      name = gtk_widget_get_name (widget);
-      if (name && !strcmp (name, "gtk-tooltips"))
-        accessible->role = ATK_ROLE_TOOL_TIP;
-      else 
-        accessible->role = ATK_ROLE_FRAME;
-    }
-  else if (GTK_IS_HANDLE_BOX (widget))
-    accessible->role = ATK_ROLE_UNKNOWN;
-  else
-    accessible->role = ATK_ROLE_INVALID;
+  /*
+   * Notify that tooltip is showing
+   */
+  if (accessible->role == ATK_ROLE_TOOL_TIP &&
+      GTK_WIDGET_MAPPED (widget))
+    atk_object_notify_state_change (accessible, ATK_STATE_SHOWING, 1);
 
   return accessible;
 }
@@ -242,14 +229,68 @@ static void
 gail_window_real_initialize (AtkObject *obj,
                              gpointer  data)
 {
+  GtkWidget *widget;
+  GailWindow *window;
+
   ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
 
+  window = GAIL_WINDOW (obj);
+  window->name_change_handler = 0;
+  window->previous_name = g_strdup (gtk_window_get_title (GTK_WINDOW (data)));
+  widget = GTK_WIDGET (data);
+
   g_signal_connect (data,
                     "window_state_event",
                     G_CALLBACK (gail_window_state_event_gtk),
                     NULL);
   g_object_set_data (G_OBJECT (obj), "atk-component-layer",
                      GINT_TO_POINTER (ATK_LAYER_WINDOW));
+
+  if (GTK_IS_FILE_SELECTION (widget))
+    obj->role = ATK_ROLE_FILE_CHOOSER;
+  else if (GTK_IS_COLOR_SELECTION_DIALOG (widget))
+    obj->role = ATK_ROLE_COLOR_CHOOSER;
+  else if (GTK_IS_FONT_SELECTION_DIALOG (widget))
+    obj->role = ATK_ROLE_FONT_CHOOSER;
+  else if (GTK_IS_MESSAGE_DIALOG (widget))
+    obj->role = ATK_ROLE_ALERT;
+  else if (GTK_IS_DIALOG (widget))
+    obj->role = ATK_ROLE_DIALOG;
+  else if (GTK_IS_WINDOW (widget))
+    {
+      const gchar *name;
+
+      name = gtk_widget_get_name (widget);
+      if (name && !strcmp (name, "gtk-tooltips"))
+        obj->role = ATK_ROLE_TOOL_TIP;
+      else if (GTK_IS_PLUG (widget))
+        obj->role = ATK_ROLE_PANEL;
+      else
+        obj->role = ATK_ROLE_FRAME;
+    }
+  else if (GTK_IS_HANDLE_BOX (widget))
+    obj->role = ATK_ROLE_UNKNOWN;
+  else
+    obj->role = ATK_ROLE_INVALID;
+}
+
+static void
+gail_window_finalize (GObject *object)
+{
+  GailWindow* window = GAIL_WINDOW (object);
+
+  if (window->name_change_handler)
+    {
+      g_source_remove (window->name_change_handler);
+      window->name_change_handler = 0;
+    }
+  if (window->previous_name)
+    {
+      g_free (window->previous_name);
+      window->previous_name = NULL;
+    }
+
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
 static G_CONST_RETURN gchar*
@@ -304,9 +345,6 @@ gail_window_get_parent (AtkObject *accessible)
 
   parent = ATK_OBJECT_CLASS (parent_class)->get_parent (accessible);
 
-  if (!parent)
-    parent = atk_get_root ();
-
   return parent;
 }
 
@@ -385,23 +423,59 @@ gail_window_ref_state_set (AtkObject *accessible)
   return state_set;
 }
 
+static gboolean
+idle_notify_name_change (gpointer data)
+{
+  GailWindow *window;
+  AtkObject *obj;
+
+  window = GAIL_WINDOW (data);
+  window->name_change_handler = 0;
+  if (GTK_ACCESSIBLE (window)->widget == NULL)
+    return FALSE;
+
+  obj = ATK_OBJECT (window);
+  if (obj->name == NULL)
+    {
+    /*
+     * The title has changed so notify a change in accessible-name
+     */
+      g_object_notify (G_OBJECT (obj), "accessible-name");
+    }
+  g_signal_emit_by_name (obj, "visible_data_changed");
+  return FALSE;
+}
+
 static void
 gail_window_real_notify_gtk (GObject		*obj,
                              GParamSpec		*pspec)
 {
   GtkWidget *widget = GTK_WIDGET (obj);
   AtkObject* atk_obj = gtk_widget_get_accessible (widget);
+  GailWindow *window = GAIL_WINDOW (atk_obj);
+  const gchar *name;
+  gboolean name_changed = FALSE;
 
   if (strcmp (pspec->name, "title") == 0)
     {
-      if (atk_obj->name == NULL)
+      name = gtk_window_get_title (GTK_WINDOW (widget));
+      if (name)
         {
-        /*
-         * The title has changed so notify a change in accessible-name
-         */
-          g_object_notify (G_OBJECT (atk_obj), "accessible-name");
-	}
-      g_signal_emit_by_name (atk_obj, "visible_data_changed");
+         if (window->previous_name == NULL ||
+             strcmp (name, window->previous_name) != 0)
+           name_changed = TRUE;
+        }
+      else if (window->previous_name != NULL)
+        name_changed = TRUE;
+
+      if (name_changed)
+        {
+          g_free (window->previous_name);
+          window->previous_name = g_strdup (name);
+       
+          if (window->name_change_handler == 0)
+            window->name_change_handler = g_idle_add (idle_notify_name_change, atk_obj);    
+        }
     }
   else
     GAIL_WIDGET_CLASS (parent_class)->notify_gtk (obj, pspec);
@@ -555,7 +629,8 @@ get_window_desktop (Window window)
                                &ret_type, &format, &nitems,
                                &bytes_after, (guchar **) &cardinals);
   error = gdk_error_trap_pop();
-  if (error != Success || result != Success)
+  /* nitems < 1 will occur if the property is not set */
+  if (error != Success || result != Success || nitems < 1)
     return -1;
 
   desktop = *cardinals;
@@ -610,7 +685,8 @@ get_stacked_windows (GailScreenInfo *info)
                                False, XA_WINDOW, &ret_type, &format, &nitems,
                                &bytes_after, (guchar **)&data);
   error = gdk_error_trap_pop ();
-  if (error != Success || result != Success)
+  /* nitems < 1 will occur if the property is not set */
+  if (error != Success || result != Success || nitems < 1)
     {
       free_screen_info (info);
       return FALSE;
